local EmojiPacks = {}
local EmojiTextureMap = {}

local INCOMPLETE_SHORTCODE = ":[^:][^:]+$"
local COMPLETE_SHORTCODE = ":[^:]+:"

Emoji = {
    SetPack = function(name, strategy)
        EmojiPacks[name] = strategy
        -- Regenerate EmojiTextureMap
        EmojiTextureMap = {}
        for packName, pack in pairs(EmojiPacks) do
            for _, emoji in pairs(pack) do
                for _, emojiName in pairs(emoji.names) do
                    EmojiTextureMap[emojiName] =
                        "\124TInterface/AddOns/" .. packName .. "/emoji/" .. emoji.file .. ":0\124t"
                end
            end
        end
    end
}

local function startsWith(str, start)
    return str:sub(1, #start) == start
end

local function oneStartsWith(table, start)
    for _, v in pairs(table) do
        if startsWith(v, start) then
            return true
        end
    end
    return false
end

local function getMatches(text, numResults)
    local results = {}
    for name, pack in pairs(EmojiPacks) do
        for _, emoji in pairs(pack) do
            if oneStartsWith(emoji.names, text) then
                table.insert(
                    results,
                    {
                        ["name"] = "\124TInterface/AddOns/" ..
                            name .. "/emoji/" .. emoji.file .. ":0\124t " .. emoji.names[1],
                        ["priority"] = LE_AUTOCOMPLETE_PRIORITY_OTHER
                    }
                )
            end
            if #results >= numResults then
                return results
            end
        end
    end
    return results
end

local function onTextChanged(self, userInput)
    if userInput == false then
        return
    end
    local text = self:GetText()
    local beginning = text:sub(1, self:GetCursorPosition())
    local shortCode = beginning:match(INCOMPLETE_SHORTCODE)
    if shortCode then
        AutoCompleteEditBox_SetAutoCompleteSource(self, getMatches)
        AutoComplete_Update(self, shortCode, #shortCode)
    end
end

local _AutoCompleteButton_OnClick = AutoCompleteButton_OnClick
function AutoCompleteButton_OnClick(self, ...)
    local autoComplete = self:GetParent()
    local editBox = autoComplete.parent
    if (editBox.autoCompleteSource == nil) then
        autoComplete:Hide()
    elseif (editBox.autoCompleteSource == getMatches) then
        local complete = self.name:match(COMPLETE_SHORTCODE .. "$")
        local cursorPosition = editBox:GetCursorPosition()
        local text = editBox:GetText()
        local beginning = text:sub(1, cursorPosition)
        local incomplete = beginning:match(INCOMPLETE_SHORTCODE)
        if (complete and incomplete) then
            local newText = beginning:sub(1, -(#incomplete) - 1) .. complete .. " "
            editBox:SetText(newText .. text:sub(cursorPosition + 1))
            editBox:SetCursorPosition(#newText)
        end
        autoComplete:Hide()
    else
        _AutoCompleteButton_OnClick(self, ...)
    end
end

for i = 1, NUM_CHAT_WINDOWS do
    local chatFrame = _G["ChatFrame" .. i .. "EditBox"]
    if chatFrame then
        chatFrame:SetScript("OnTextChanged", onTextChanged)
    end
end

local CHAT_MSG_TYPES = {
    "AFK",
    "BATTLEGROUND_LEADER",
    "BATTLEGROUND",
    "CHANNEL",
    "DND",
    "EMOTE",
    "GUILD",
    "OFFICER",
    "PARTY_LEADER",
    "PARTY",
    "RAID_LEADER",
    "RAID_WARNING",
    "RAID",
    "SAY",
    "WHISPER",
    "WHISPER_INFORM",
    "YELL"
}

function onChatMessage(self, event, message, ...)
    local links = {}
    message =
        message:gsub(
        "\124H.-\124h.-\124h",
        function(link)
            table.insert(links, link)
            return ("\2" .. #links .. "\3")
        end
    )
    message =
        message:gsub(
        COMPLETE_SHORTCODE,
        function(shortCode)
            local texture = EmojiTextureMap[shortCode]
            if (texture) then
                return texture
            else
                return shortCode
            end
        end
    )
    if #links > 0 then
        for index, link in pairs(links) do
            message = message:gsub("\2" .. index .. "\3", link)
        end
    end
    return false, message, ...
end

for _, type in pairs(CHAT_MSG_TYPES) do
    ChatFrame_AddMessageEventFilter("CHAT_MSG_" .. type, onChatMessage)
end